/****************************************************************************
  This program samples images at set scales and smoothens them as a function of the scale.

  Format: create_sampled_data_base [image]
  
  Input: [image] = image_name.pgm

  Author: Nikhil V Shirahatti
  Date: 11/24/03
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

extern int read_pgm_hdr(FILE *, int *, int *);
extern void **matrix(int, int, int, int, int);
extern void error(const char *);


/*/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*/

/* READ PGM HEADER */

/* This function reads the header of a PGM image. */
/* The dimensions are returned as arguments. */
/* This function ensures that there's no more than 8 bpp. */
/* The return value is negative if there's an error. */

int read_pgm_hdr(FILE *fp, int *nrows, int *ncols)
{
  char filetype[3];
  int maxval;

  if(skipcomment(fp) == EOF
     || fscanf(fp, "%2s", filetype) != 1
     || strcmp(filetype, "P5")
     || skipcomment(fp) == EOF
     || fscanf(fp, "%d", ncols) != 1
     || skipcomment(fp) == EOF
     || fscanf(fp, "%d", nrows) != 1
     || skipcomment(fp) == EOF
     || fscanf(fp, "%d%*c", &maxval) != 1
     || maxval > 255)
    return(-1);
  else return(0);
}

/*-----------------------------------------------------------------------------------*/
/* ERROR HANDLER */

void error(const char *msg)
{
  fprintf(stderr, "%s\n", msg);
  exit(1);
}

/*------------------------------------------------------------------------------------*/

/* DYNAMICALLY ALLOCATE A PSEUDO 2-D ARRAY */

/* This function allocates a pseudo 2-D array of size nrows x ncols. */
/* The coordinates of the first pixel will be first_row_coord and */
/* first_col_coord. The data structure consists of one contiguous */
/* chunk of memory, consisting of a list of row pointers, followed */
/* by the array element values. */
/* Assumption:  nrows*ncols*element_size, rounded up to a multiple  */
/* of sizeof(long double), must fit in a long type.  If not, then */
/* the "i += ..." step could overflow.                              */

void **matrix(int nrows, int ncols, int first_row_coord,
              int first_col_coord, int element_size)
{
  void **p;
  int alignment;
  long i;

  if(nrows < 1 || ncols < 1) return(NULL);
  i = nrows*sizeof(void *);
  /* align the addr of the data to be a multiple of sizeof(long double) */
  alignment = i % sizeof(long double);
  if(alignment != 0) alignment = sizeof(long double) - alignment;
  i += nrows*ncols*element_size+alignment;
  if((p = (void **)malloc((size_t)i)) != NULL)
  {
    /* compute the address of matrix[first_row_coord][0] */
    p[0] = (char *)(p+nrows)+alignment-first_col_coord*element_size;
    for(i = 1; i < nrows; i++)
      /* compute the address of matrix[first_row_coord+i][0] */
      p[i] = (char *)(p[i-1])+ncols*element_size;
    /* compute the address of matrix[0][0] */
    p -= first_row_coord;
  }
  return(p);
}

/*----------------------------------------------------------------------------------------------------------------------*/
/* SKIP COMMENT */

/* This function skips past a comment in a file. The comment */
/* begins with a '#' character and ends with a newline character. */
/* The function returns EOF if there's an error. */

int skipcomment(FILE *fp)
{
  int i;

  if((i = getc(fp)) == '#')
    while((i = getc(fp)) != '\n' && i != EOF);
  return(ungetc(i, fp));
}


/*---------------------------------------------------------------------------------------------------------------------*/

/* REFLECT AN IMAGE ACROSS ITS BORDERS */

/* The parameter "amount" tells the number of rows or columns to be */
/* reflected across each of the borders. */
/* It is assumed that the data type is unsigned char. */
/* It is assumed that the array was allocated to be of size at least */
/* (nrows+2*amount) by (ncols+2*amount), and that the image was loaded */
/* into the middle portion of the array, with coordinates, */
/* 	0 <= row < nrows, 0 <= col < ncols */
/* thereby leaving empty elements along the borders outside the image */
/* The "reflect" function will then fill in those empty */
/* elements along the borders with the reflected image pixel values. */
/* For example, x[0][-1] will be assigned the value of x[0][0], */
/* and x[0][-2] will be assigned the value of x[0][1], if amount=2. */

void reflect(unsigned char **xc, int nrows, int ncols, int amount)
{
  int i, j;

  if(matrix == NULL || nrows < 1 || ncols < 1 || amount < 1
    || amount > (nrows+1)/2 || amount > (ncols+1)/2)
    error("reflect: bad args");

  for(i = -amount; i < 0; i++)
  {
    for(j = -amount; j < 0; j++)
      xc[i][j] = xc[-i-1][-j-1];
    for(j = 0; j < ncols; j++)
      xc[i][j] = xc[-i-1][j];
    for(j = ncols; j < ncols+amount; j++)
      xc[i][j] = xc[-i-1][ncols+ncols-j-1];
  }
  for(i = 0; i < nrows; i++)
  {
    for(j = -amount; j < 0; j++)
      xc[i][j] = xc[i][-j-1];
    for(j = ncols; j < ncols+amount; j++)
      xc[i][j] = xc[i][ncols+ncols-j-1];
  }
  for(i = nrows; i < nrows+amount; i++)
  {
    for(j = -amount; j < 0; j++)
      xc[i][j] = xc[nrows+nrows-i-1][-j-1];
    for(j = 0; j < ncols; j++)
      xc[i][j] = xc[nrows+nrows-i-1][j];
    for(j = ncols; j < ncols+amount; j++)
      xc[i][j] = xc[nrows+nrows-i-1][ncols+ncols-j-1];
  }
}
/*---------------------------------------------------------------------------------------------*/

/* REFLECTING FLOAT */

void reflectf(float **xc, int nrows, int ncols, int amount)
{
 int i, j;

  if(matrix == NULL || nrows < 1 || ncols < 1 || amount < 1
    || amount > (nrows+1)/2 || amount > (ncols+1)/2)
    error("reflect: bad args");

  
  for(i = -amount; i < 0; i++)
  {
    for(j = -amount; j < 0; j++)
      xc[i][j] = xc[-i-1][-j-1];
    for(j = 0; j < ncols; j++)
      xc[i][j] = xc[-i-1][j];
    for(j = ncols; j < ncols+amount; j++)
      xc[i][j] = xc[-i-1][ncols+ncols-j-1];
  }
  
  for(i = 0; i < nrows; i++)
  {
    for(j = -amount; j < 0; j++)
      xc[i][j] = xc[i][-j-1];
    for(j = ncols; j < ncols+amount; j++)
      xc[i][j] = xc[i][ncols+ncols-j-1];
  }
  
  for(i = nrows; i < nrows+amount; i++)
  {
    for(j = -amount; j < 0; j++)
      xc[i][j] = xc[nrows+nrows-i-1][-j-1];
    for(j = 0; j < ncols; j++)
      xc[i][j] = xc[nrows+nrows-i-1][j];
    for(j = ncols; j < ncols+amount; j++)
      xc[i][j] = xc[nrows+nrows-i-1][ncols+ncols-j-1];
  }
}
/*-------------------------------------------------------------------------------------------------------------------*/


/* ====================================================================================*/

int main(int argc, char **argv)
{
  
  FILE *fpx, *fpy;
  int nrows, ncols, i, j,thresh,thresh2;
  unsigned char **x;
  char *str;
  int r=0,p=0,s=0,f=0,histh=0;
  
  /* PARSE THE COMMAND LINE */
  while(--argc > 0 && **(argv+1) == '-')
    switch((*++argv)[1])
    {
            
    default:
      error("Usage:edge {-roberts,-prewitt,-sobel,-frei} [-skipNMS] [-t thresh_low thresh_high] img ");
      break;


    }   /* end of case statement */

  /* OPEN FILES */
  printf("------------------------------------------------\n");
  printf("Opening image file\n"); 
  if(argc == 0) fpx = stdin;
  else 
   {
     printf(" Making decesion\n");
     if(histh)
	{	
	  if((fpx = fopen(*(++argv), "r")) == NULL)
	  {    		
      		  printf("%s1\n",(*argv));
      		  error("can't open file");
	  }
	}		
	else
	{
	  if((fpx = fopen((str), "r")) == NULL)
	  {  
	     printf("%s2\n",(*argv));
	     error("can't open file");
	  }
        }
   }    
  printf("-----------------------------------------\n");
  printf(" Opened file --image file %s \n", *argv);
  fpy = stdout;

  /* READ HEADER */
  if(read_pgm_hdr(fpx, &nrows, &ncols) < 0)
    error("not a PGM image or bpp > 8");
  printf("------------------------------------------\n");
  printf(" Read Header \n");


  /* ALLOCATE ARRAYS */
  /* The x array is extended by one pixel on all four sides. */
  /* The first element of the array is x[-1][-1]. */
  /* The first actual pixel of the image will be stored at x[0][0]. */
  /* y is the output image -- magnitude */
  /* z is the output image -- gradient  */
  /* edge_map is the matrix containbing whether a given pixel is an edge or not */

  x = (unsigned char **)matrix(nrows+2, ncols+2, -1, -1, sizeof(char));  
  printf("--------------------------------------------------------\n");
  printf("Allocated arrays for output image:gradient and magnitude and edge map\n");
  if(x == NULL) error("can't allocate memory");


  /* READ THE IMAGE */
  for(i = 0; i < nrows; i++)
    if(fread(&x[i][0], sizeof(char), ncols, fpx) != ncols)
      error("can't read the image");
  printf("----------------------------------------------\n");
  printf(" Read image: %s\n",*argv);

  /* REFLECT THE IMAGE ACROSS ITS BORDERS BY ONE PIXEL */
  reflect(x, nrows, ncols, 1);
  printf("-------------------------------------------------\n");
  printf("Finished processing: Reflected image by 1 pixel\n");
  printf("TEST %d \n",x[-1][-1]);

  
  fclose(fpx);
  return 1;
}

/*/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*/
